#ifndef __SpiceController_H
#define __SpiceController_H

/*
    Basic assumption:
    ------------------
    Cross platform compatible.
    Easy to transform into remote process communication
    Secured

    choosen:
        Unix - Unix Domain Sockets (easy to change into regular sockets for remote communication)
        Windows - Named pipe (which allows remote access and is duplex
            (rather than anonymous pipe which is local only and one way)
*/

#include "controller_messages.h"
#include "application.h"
#include "spice_configuration.h"

class Application;

#define CONTROLLER_BUFFER_SIZE 4096

enum {
    MSG_CONNECT,
    MSG_SHOW,
    MSG_host,
    MSG_port,
    MSG_sport,
    MSG_full_screen,
    MSG_init_ipc,
    MSG_ticketing,
    MSG_no_audio,
    MSG_auto_display_res,
    MSG_password,
    MSG_cipher,
    MSG_title,
    MSG_usb_server_port,
    MSG_encrypt_channels,
    MSG_hotkeys,
    MSG_send_cad,
#ifdef WIN32
    MSG_no_taskmgr,
#endif
    MSG_guest_host_name,
    MSG_CA_file,
    MSG_host_cert_subject,
    MSG_END
};

typedef struct ControlMessage {
    int id;
    char szData[CONTROLLER_BUFFER_SIZE];
} ControlMessage;

class SpiceController {
    public:
    SpiceController (SpiceConfiguration *config);
    SpiceController (MessageData &data);

    ~SpiceController();

    void Create ();
    int Connect ();
    int Connect (int nRetries);
    void Disconnect();
    void DisconnectClient();

    void Write (
        const void *lpBuffer,
        uint32_t nBytesToWrite,
        uint32_t *nBytesWritten
    );
    void Read (
        void *lpBuffer,
        uint32_t nBytesToRead,
        uint32_t *nBytesRead
    );

    bool new_client_connection();
    bool WaitForPipe (uint32_t nTimeout);
    bool WaitForConnectMsg (uint32_t nTimeout);

    MessageData* get_data() { return m_data; };

    void ProcessMessage (ControlMessage& msg);
        int TranslateRC(int nRC);

    void SetApp (Application *app);
    SpiceConfiguration *GetConfig();


    private:
    MessageData *m_data;

    wchar_t wcPipeName[256];
    uint32_t nOpenMode;
    uint32_t nPipeMode;
    uint32_t nMaxInstances;
    uint32_t nOutBufferSize;
    uint32_t nInBufferSize;

    int nReadPos;
    int nCurrentReadLength;

    ControlMessage msg;
    Application *_app;
    SpiceConfiguration *_spice_config;

    bool _is_ipc_connect_recieved;


};

#endif
