#ifndef _H_FOREIGN_MENU
#define _H_FOREIGN_MENU

#include <map>
#include <boost/function.hpp>
#include "named_pipe.h"

class ForeignMenuItem;
struct FrgMenuMsg;

class ForeignMenuInterface {
public:
    virtual void on_foreign_menu_message(FrgMenuMsg *hdr, int32_t opaque_conn_ref) = 0;
    virtual void delete_menu(int32_t opaque_conn_ref) = 0;
    virtual ~ForeignMenuInterface() {}
};

class ForeignMenu : public NamedPipe::ListenerInterface {
public:
    ForeignMenu(ForeignMenuInterface *handler);
    virtual ~ForeignMenu();
    virtual NamedPipe::ConnectionInterface &create();
    void add_item_connection(NamedPipe::ConnectionRef conn_ref, ForeignMenuItem *menu_item);
    void on_menu_activated();
    void on_menu_deactivated();
    void on_foreign_menu_command(NamedPipe::ConnectionRef conn_ref, int32_t id);
    void item_connection_closed(NamedPipe::ConnectionRef conn_ref);

private:
    void handle_menu_activated(ForeignMenuItem *item, int32_t cmd);
    void send_msg_to_client(NamedPipe::ConnectionRef conn_ref, FrgMenuMsg *msg);
    void send_msg_to_client(ForeignMenuItem *item, FrgMenuMsg *msg);

private:
    ForeignMenuInterface *_handler;
    std::map<NamedPipe::ConnectionRef, ForeignMenuItem*> _connections;
    NamedPipe::ListenerRef _foreign_menu;
    bool _active;
};

#define FOREIGN_MENU_BUF_SIZE   4096

class ForeignMenuItem : public NamedPipe::ConnectionInterface {
public:
    ForeignMenuItem(ForeignMenuInterface *handler, ForeignMenu& parent);
	virtual ~ForeignMenuItem();
    virtual void bind(NamedPipe::ConnectionRef conn_ref);
    virtual void on_data();
    void write_msg(const void *buf, int len);
    bool read_msgs();

private:
    void handle_foreign_menu_message(FrgMenuMsg *hdr, NamedPipe::ConnectionRef conn_ref);

private:
    ForeignMenuInterface *_handler;
	ForeignMenu& _parent;
    int _write_pending;
    uint8_t *_write_pos;
    uint8_t *_read_pos;
    uint8_t _write_buf[FOREIGN_MENU_BUF_SIZE];
    uint8_t _read_buf[FOREIGN_MENU_BUF_SIZE];
};

#endif
