#ifndef _H_REDCLIENTCONFIGURATION
#define _H_REDCLIENTCONFIGURATION

#include <string>

enum {
    SERVER_AUTH_PUBKEY = 1,
    SERVER_AUTH_NAME = (1 << 1),
    SERVER_AUTH_SUBJECT = (1 << 2),
};

typedef std::pair<std::string, std::string> CertFieldValuePair;
typedef std::list<CertFieldValuePair> CertFieldValueList;

class peer_connect_options {
public:
    peer_connect_options();
	~peer_connect_options();

    peer_connect_options(const peer_connect_options& other);

	void SetSSLParameters (
		bool bIsSSL,
		const char *szCertFile,
		const char *szKeyFile,
		const char *szPassword,
		const char *szChipers,
		const uint16_t nPort
	);

    void SetSSLServerAuthParameters (
        int nServerAuthFlags,
        const uint8_t *pPubKey,
        int nPubKeySize,
        const char *szCAFile,
        const CertFieldValueList& lSubject
    );
	
    void assign(const peer_connect_options& other);
    bool IsSSL() const {return _bIsSSL;}
    const char *GetCertFile() const { return strlen(_szCertFile) ? _szCertFile : NULL;}
    const char *GetKeyFile() const { return strlen(_szKeyFile) ? _szKeyFile : NULL;}
    const char *GetPassword() const { return strlen(_szPassword) ? _szPassword : NULL;}
    int GetServerAuthFlags() const { return _nServerAuthFlags;}
    const char *GetCAFile() const { return strlen(_szCAFile) ? _szCAFile : NULL;}
    const char *GetCipher() const { return strlen(_szCipher) ? _szCipher : NULL;}
    const uint8_t* GetPubKey() const { return _pPubKey;}
    int GetPubKeySize() const { return _nPubKeySize;}
    const CertFieldValueList& GetSubject() const { return _lSubject;}
    uint16_t GetPort() const { return _nPort;}
	
private:
	bool _bIsSSL;
	char _szCertFile[256];
	char _szKeyFile[256];
	char _szPassword[256];
	char _szCipher[256];
    int _nServerAuthFlags;
	char _szCAFile[256];
    uint8_t* _pPubKey;
    int _nPubKeySize;
    CertFieldValueList _lSubject;

	uint16_t _nPort;
};

#endif

