#ifndef _H_READ_WRITE_MUTEX
#define _H_READ_WRITE_MUTEX
/* inspired by boost shared_mutex */

// Support for old boost::thread
#include <boost/thread/mutex.hpp>
#ifndef BOOST_THREAD_MUTEX_HPP
#include <boost/thread/detail/lock.hpp>
typedef boost::detail::thread::scoped_lock<boost::mutex> Lock;
#else
typedef boost::mutex::scoped_lock Lock;
#endif
#include <boost/thread/condition.hpp>

class ReadWriteMutex {
public: 
    ReadWriteMutex()
    {
        _state.read_count = 0;
        _state.write = false;
        _state.write_waiting = false;
    }

    virtual ~ReadWriteMutex()
    {
    }

    void read_lock()
    {
        Lock lock(_state_mutex);

        while (_state.write || _state.write_waiting)
        {
            _read_cond.wait(lock);
        }

        ++_state.read_count;
    }

    bool try_read_lock()
    {
        Lock lock(_state_mutex);
        if (_state.write || _state.write_waiting)
        {
            return false;
        }
        else 
        {
            ++_state.read_count;
            return true;
        }
    }

    void read_unlock()
    {
        Lock lock(_state_mutex);
        --_state.read_count;

        if (!_state.read_count) // last reader
        {
            _state.write_waiting = false; 
            release_waiters();
        }
    }

    void write_lock()
    {
        Lock lock(_state_mutex);

        while (_state.read_count || _state.write) 
        {
            _state.write_waiting = true;
            _write_cond.wait(lock);
        }
        _state.write = true;
    }

    bool try_write_lock()
    {
        Lock lock(_state_mutex);

        if (_state.read_count || _state.write)
        {
            return false;
        }
        else
        {
            _state.write = true;
            return true;
        }
    }

    void write_unlock()
    {
        Lock lock(_state_mutex);
        _state.write = false;
        _state.write_waiting = false;
        release_waiters();
    }

private:
    void release_waiters()
    {
        _write_cond.notify_one();
        _read_cond.notify_all();
    }
private:
    struct {
        unsigned int read_count;
        bool         write;
        bool         write_waiting;
    } _state;

    boost::mutex      _state_mutex;
    boost::condition  _read_cond;
    boost::condition  _write_cond;
};

#endif
