#ifndef _H_REDPEER
#define _H_REDPEER

#ifdef _WIN32
#include <winsock.h>
#else
typedef int SOCKET;
#endif

#include <openssl/ssl.h>
#include <openssl/err.h>

#include "common.h"
#include "red.h"
#include "peer_connect_options.h"
#include "events_loop.h"

class RedPeer : protected EventsLoop::Socket {
public:
    RedPeer();
    virtual ~RedPeer();

    class InMessage;
    class CompundInMessage;
    class OutMessage;
    class DisconnectedException {};

    void connect(peer_connect_options& configuration, const char* host, uint16_t port);
    void disconnect();
    void swap(RedPeer* other);
    void close();
    void enable() { _shut = false;}

    virtual CompundInMessage* recive();
    uint32_t send(OutMessage& message);

    uint32_t recive(uint8_t* buf, uint32_t size);
    uint32_t send(uint8_t* buf, uint32_t size);

    static uint32_t host_by_name(const char *host);

protected:
    virtual void on_event() {}
    virtual int get_socket() { return _peer;}

    static bool verify_pubkey(X509* cert, const uint8_t* key, int key_len);
    static bool verify_host_name(X509* cert, const char* host_name);
    static bool verify_subject(X509* cert, const CertFieldValueList& subject);

    static int ssl_verify_callback(int preverify_ok, X509_STORE_CTX *ctx);

    static bool x509_cert_host_name_compare(const char *cert_name, int cert_name_size,
                                            const char *host_name);

private:
    SOCKET _peer;
    Mutex _lock;
    bool _shut;
    uint64_t _serial;
    
    SSL_CTX *_ctx;
    SSL *_ssl;
    BIO *_sbio;
};

class RedPeer::InMessage {
public:
    InMessage(uint16_t type, uint32_t size, uint8_t* data)
        : _type (type)
        , _size (size)
        , _data (data) 
    {
    }

    virtual ~InMessage() {}
    
    uint16_t type() { return _type;}
    uint32_t size() { return _size;}
    uint8_t* data() { return _data;}

protected:
    uint16_t _type;
    uint32_t _size;
    uint8_t* _data;
};

class RedPeer::CompundInMessage : public RedPeer::InMessage {
public:
    CompundInMessage(uint64_t _serial, uint16_t type, uint32_t size, uint32_t sub_list)
        : InMessage(type, size, new uint8_t[size])
        , _serial (_serial)
        , _sub_list (sub_list)
    {
    }

    virtual ~CompundInMessage() { delete[] _data;}
    
    uint64_t serial() { return _serial;}
    uint32_t sub_list() { return _sub_list;}

private:
    uint64_t _serial;
    uint32_t _sub_list;
};

class RedPeer::OutMessage {
public:
    OutMessage(uint32_t type, uint32_t size);
    virtual ~OutMessage();

    RedDataHeader& header() { return *(RedDataHeader *)_data;}
    uint8_t* data() { return _data + sizeof(RedDataHeader);}
    void resize(uint32_t size);

private:
    uint32_t message_size() { return header().size + sizeof(RedDataHeader);}
    uint8_t* base() { return _data;}

private:
    uint8_t* _data;
    uint32_t _size;

    friend class RedPeer;
    friend class RedChannel;
};

#endif

