#ifndef _H_SCREEN_LAYER
#define _H_SCREEN_LAYER

#include "region.h"

class RedScreen;
class Application;

class ScreenLayer {
public:
    ScreenLayer(int z_order, bool opaque);
    virtual ~ScreenLayer();

    void attach_to_screen(Application& applicaion, int screen_id);
    void detach_from_screen(Application& applicaion);

    void set_screen(RedScreen* screen) { _screen = screen;}
    RedScreen* screen() { return _screen; }
    int z_order() { return _z_order;}
    void set_area(const QRegion& area);
    void offset_area(int dx, int dy);
    void clear_area();
    void set_rect_area(const Rect& r);
    void add_rect_area(const Rect& r);
    void remove_rect_area(const Rect& r);
    void begin_update(QRegion& direct_rgn, QRegion& composit_rgn);
    void invalidate();
    uint64_t invalidate(const Rect& r, bool urgent = false);
    void invalidate(const QRegion& r);

    virtual void copy_pixels(const QRegion& dest_region, RedDrawable& dest_dc) {}
    virtual void copy_pixels(const QRegion& dest_region, const PixmapHeader &dest, RedGLContext helper) {}
    virtual void copy_pixels(const QRegion& dest_region, RedDrawable& dest_dc, RedGLContext ogl_context) {}
    
    void set_using_ogl(bool val) {_using_ogl = val;}
    bool using_ogl() {return _using_ogl;}
    virtual void on_size_changed() {}

    virtual void pre_migrate() { }
    virtual void post_migrate() { }

    QRegion& area() { return _area;}
    QRegion& direct_area() { return _direct_area;}
    QRegion& composit_area() { return _composit_area;}

    virtual void on_update_completion(uint64_t mark) {}

private:
    uint64_t invalidate_rect(const Rect& r, bool urgent);

private:
    RedScreen* _screen;
    int _z_order;
    bool _opaque;
    bool _using_ogl;
    Mutex _area_lock;
    QRegion _area;
    QRegion _direct_area;
    QRegion _composit_area;
};

#endif
