#ifndef __spice_configuration_h
#define __spice_configuration_h

#include <map>

#include "red.h"
#include "redkey.h"
#include "debug.h"
#include "peer_connect_options.h"

class SpiceConfiguration {
public:

    SpiceConfiguration();

    struct hotkey_t {
        RedKey main;
        RedKey alter;
    };

    typedef std::vector<hotkey_t> hotkey_seq_t;
    typedef std::map<int, hotkey_seq_t> hotkeys_t;

    enum {
        ActionUnknown = -1,
        ActionToggleFullScreen,
        ActionReleaseCursor,
#ifdef RED_DEBUG
        ActionConnect,
        ActionDisconnect,
#endif// RED_DEBUG
        ActionLast
    };

public:

    void parse_hotkeys(const std::string& hotkeys);

    /* Subject entries are separated by ','. 
      entry: <attribute-name>=<attribute-value>
      If ',' appears inside <attribute-value>, it should be preceeded by '\'.
      If '\' appears inside <attribute-value>, it should be preceeded by '\'. 
    */
    void parse_host_cert_subject(const std::string& subject);

public:
    std::string _host;
    uint16_t _port;
    uint16_t _sport;
    bool _full_screen;
    bool _init_controller;
    bool _no_audio;
    bool _auto_display_res;
    std::string _password;
    std::string _cipher;
    std::wstring _title;
    std::string _logger_ini;
    bool _encrypt_channels[RED_CHANNEL_END];
    hotkeys_t _hotkeys;
    bool _send_cad;
#ifdef WIN32
    bool _no_taskmgr;
#endif
    std::string _canvas_type;
    std::string _CA_file;
    CertFieldValueList _host_cert_subject;

private:

    void parse_keys(int action, std::string& hotkey);
    int action_name(const std::string& action);
    void add_hotkey(const std::string& hotkey);
    void add_hotkey_redkey(std::vector<hotkey_t>& keys, const char* key);
};

#endif
