#include "common.h"
#include "utils.h"

std::wstring char2wstring (const char *src) {
    int i = mbstowcs (NULL, src, strlen(src)); 

    if (i >= 0) {
        std::vector<wchar_t> dest(i+1);
 
        i = mbstowcs(&dest[0], src, strlen(src));

        dest[i] = L'\0';
        return std::wstring(&dest[0]);
    } else {
        return std::wstring(L"");
    }
}

#ifdef WIN32
HDC create_compatible_dc()
{
    HDC dc = CreateCompatibleDC(NULL);
    if (!dc) {
        throw Exception(fmt("%s: create compatible DC failed") % __FUNCTION__ );
    }
    return dc;
}

HBITMAP get_bitmap_res(int id)
{
    HBITMAP bitmap = LoadBitmap(GetModuleHandle(NULL), MAKEINTRESOURCE(id));
    if (!bitmap) {
        throw Exception(fmt("%s: get bitmpa #%d failed") % __FUNCTION__ % id);
    }
    return bitmap;
}

HBITMAP get_alpha_bitmap_res(int id)
{
    AutoGDIObject bitmap(LoadImage(GetModuleHandle(NULL), MAKEINTRESOURCE(id), IMAGE_BITMAP, 0, 0,
                         LR_DEFAULTCOLOR | LR_CREATEDIBSECTION | LR_SHARED));
    if (!bitmap.valid()) {
        throw Exception(fmt("%s: get bitmpa #%d failed") % __FUNCTION__ % id);
    }

    BITMAP src_info;
    GetObject(bitmap.get(), sizeof(src_info), &src_info);
    if (src_info.bmBitsPixel != 32 || src_info.bmPlanes != 1) {
        throw Exception(fmt("%s: invalid format #%d ") % __FUNCTION__ % id);
    }

    LONG src_size = src_info.bmHeight * src_info.bmWidthBytes;
    boost::scoped_array<uint8_t> src_pixels(new uint8_t[src_size]);
    LONG ncopy = GetBitmapBits((HBITMAP)bitmap.get(), src_size, src_pixels.get());
    if (ncopy != src_size) {
        throw Exception(fmt("%s: get vitmap bits failed, %u") % __FUNCTION__ % GetLastError());
    }

    AutoDC auto_dc(create_compatible_dc());
    BITMAPINFO dest_info;
    uint8_t *dest;
    dest_info.bmiHeader.biSize = sizeof(dest_info.bmiHeader);
    dest_info.bmiHeader.biWidth = src_info.bmWidth;
    dest_info.bmiHeader.biHeight = -src_info.bmHeight;
    dest_info.bmiHeader.biPlanes = 1;
    dest_info.bmiHeader.biBitCount = 32;
    dest_info.bmiHeader.biCompression = BI_RGB;
    dest_info.bmiHeader.biSizeImage = 0;
    dest_info.bmiHeader.biXPelsPerMeter = dest_info.bmiHeader.biYPelsPerMeter = 0;
    dest_info.bmiHeader.biClrUsed = 0;
    dest_info.bmiHeader.biClrImportant = 0;

    HBITMAP ret = CreateDIBSection(auto_dc.get(), &dest_info, 0, (VOID **)&dest, NULL, 0);
    if (!ret) {
        throw Exception(fmt("%s: create bitmap failed, %u") % __FUNCTION__ % GetLastError());
    }

    uint8_t* src_line = src_pixels.get();
    for (int i = 0; i < src_info.bmHeight; i++, src_line += src_info.bmWidthBytes) {
        uint8_t* src = src_line;
        for (int j = 0; j < src_info.bmWidth; j++) {
            dest[3] = src[3];
            double alpha = (double)dest[3] / 0xff;
            dest[2] = (uint8_t)(alpha * src[2]);
            dest[1] = (uint8_t)(alpha * src[1]);
            dest[0] = (uint8_t)(alpha * src[0]);
            src += 4;
            dest += 4;
        }
    }
    return ret;
}

#endif

