#include "common.h"
#include <fstream>
#include "command_line.h"
#include "spice_configuration.h"
#include "application.h"

#include <log4cpp/PropertyConfigurator.hh>
#include <log4cpp/BasicConfigurator.hh>

const char* const LOGGER_CONFIG = "logger.ini";

static void cleanup()
{
    log4cpp::Category::shutdown();
}

int main(int argc, const char** argv)
{
    int exit_val;

    atexit(cleanup);
    try {
        LOGGER_SECTION("spice");

        std::ifstream log_init(LOGGER_CONFIG);
        if (log_init.good()) {
            log_init.close();
            log4cpp::PropertyConfigurator::configure(LOGGER_CONFIG);
        } else {
            log4cpp::BasicConfigurator::configure();
        }

        SpiceConfiguration spice_config;
        CommandLine cmd(spice_config);

        DBG (0, "spicec args:");
        for (int i=0; i < argc; i++) {
            DBG(0, fmt("\t%s") % argv[i]);
        }
        DBG(0, "----------- End of spicec args ---------------");

        if (!cmd.parse_cmd_line(argc, argv)) {
            return SPICEC_ERROR_CODE_ERROR;
        }

        log_init.open(cmd.get_logger_ini().c_str());
        if (log_init.good()) {
            LOG_DEBUG("reconfiguring logger");
            log_init.close();
            log4cpp::PropertyConfigurator::configure(cmd.get_logger_ini().c_str());
        }

        Application::init();
        std::auto_ptr<Application> app(new Application());
        app->config(spice_config);
        app->set_send_cad(true);
        exit_val = app->run();
    } catch (Exception& e) {
        WARN(fmt("unhandle exception: %s") % e.what());
        exit_val = e.get_error_code();
    } catch (std::exception& e) {
        WARN(fmt("unhandle exception: %s") % e.what());
        exit_val = SPICEC_ERROR_CODE_ERROR;
    } catch (...) {
        WARN("unhandled exception");
        exit_val = SPICEC_ERROR_CODE_ERROR;
    }

    return exit_val;
}

